
/*
Realisiert mit AT-Mega8

Ausgnge:
PB1     Links
PB2     Rechts
PB3     Triac Motor
PB4     Schnell
PB5     Heizung
PC5     Pumpe
PD1     Tx + Watchdog
PD3     Tr (J7)
PD5     Ventil1
PD6     Ventil2
PD7     Ventil3

Eingnge:
PB0     Druckdose
PC0     Temperatur
PC1     KeinDruck
PC2     Umotor
PC3     Spannung J7 (Tr)
PC4     Netzspannung
PC6     Reset
PD0     RX
PD2     Nulldurchgang
PD4     Tacho


==========================================
T0 - 8Bit:  zhlt Motorpulse (Pin T0, PD4). Auswertung nach 100ms in T2-Interrupthandler
T1 - 16Bit: zhlt mit prescaler=8. Mit Input-Capture Unit. Druckdose auf ICP1(PB0).
T2 - 8bit:  250s Zeitbasis (synchron zum Nulldurchgang)

INT0: Nulldurchgang (50Hz)  (PD2)
ICP1: Druckdose (34..45Hz)  (PB0)

Phasenanschnittsteuerung des Motor-Triacs ber 4KHz ISR  (PB3)
==========================================



Druckdose: 
  Takt=4MHz
  Prescaler=8
  Zhlfrequenz:500KHz an Timer T1 (16bit)
  bei 45Hz: 11111
  bei 34 Hz:14705
  Differenz:3594 Schritte
  Ausgabe: Zhler (frher: -10000)
  
  
Motordrehzahl:
    Drehgeber gibt 8 Pulse pro Umdrehung (evtl. auch 10)
    Bei langsamster Drehzahl: 50Hz  (5  1/s, 300 1/min Trommel dreht dann mit 30 1/min)
    Bei max. Drehzahl: 15000 1/min = 250 1/s > 2500Hz
    Auswertung alle 100ms - Zhler 5..250
    
    Regelung:
        I-Regler: Vorzeichen der Differenz aus Uist und Usoll geht auf Integrator. 
        Integratorausgang geht auf Phasenanschnitt.
    
   

Istwerte:
    alle 100ms:
        Motordrehzahl in 1/s
                            
    
    alle 500ms:
        Druckdose          
        Thermometer       
        Umotor              
        Upumpe
        KeinDruck          
        Tr                
*/



#define F_CPU 4000000UL	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdint.h>
//#include <string.h>

#include "itoa.c"








//Der Einfachheit halber: viele globale Variablen

//Wir haben insgesamt 512 Bytes SRAM
typedef struct {
     uint8_t txMsg[80]; 
     uint8_t rxbuffer[10];
     uint8_t rxZeile[10];
     int8_t rxIndexRead;
     int8_t rxIndexStore;

     uint8_t tic; //1ms Takt - wird bei jeder Vollwelle im Nulldurchgang auf 0 gesetzt
     uint8_t sec; //0..59
     uint8_t acAdcVal[5];
     uint8_t acAdcVal2[5];
     uint8_t cAdcCh;
    
     int16_t iDruckdose;
     uint8_t cZuendzeit; //0..38 (Motortriac Zndzeitpunkt)
     uint8_t ms10;       //10ms Zhler
     uint8_t hz10;      //100ms Indikator
     uint8_t cMotorSoll; //Sollgeschwindigkeit
     uint8_t cMotorIst;  //istgeschwindigkeit in Umdrehungen/s
     uint8_t cReglersteigung;  //Increment von cZuendzeit
     uint8_t next100ms;        // wird alle 100ms auf 1 gesetzt
     char    empfangszeichen;
     char    talk;
     char    cRegelung;    //Motordrehzahl: 0=Regler AUS, 1=Regler EIN
    
}Tglob;

//Alle globalen Var in einer Struct deklarieren
Tglob g;

unsigned char *p; //Hilfsvariable fr das zusammenbauen von Strings mit eigenem itoa

//--------------------------------------------------------
//--------------------------------------------------------
ISR(INT0_vect){ //Nulldurchgang 50Hz Interrupt
    //Ticker synchronisieren
    SFIOR = _BV(PSR2); //clear Timer 2 prescaler
    
    g.tic = 0;
    OCR2 = 62;  //Mitte von 125
}


//--------------------------------------------------------
ISR(TIMER2_COMP_vect){ //4KHz Interrupt

   if (0==g.tic){ //Nulldurchgang  - Anfang Vollwelle
        //ADC auslesen und neu starten (ohne Kanalwechsel, free running)
        g.acAdcVal[g.cAdcCh] = ADCH;
//        ADMUX=0x60 | g.cAdcCh; //RefS1=0 RefS0=1 (Uref=AVcc) ADLAR=1
        ADCSRA=0xE4;  //ADEN ADSC ADFR (ADIF)   (ADIE) ADPS2..0=1.0.0  (Division Factor=16)
   }else
   if (20==g.tic){
        //ADC auslesen, Kanalwechsel und neu starten (free running)
        g.acAdcVal2[g.cAdcCh] = ADCH;
        g.cAdcCh++;
        if (g.cAdcCh > 4){
          g.cAdcCh = 0;
        }
        ADMUX=0x60 | g.cAdcCh; //RefS1=0 RefS0=1 (Uref=AVcc) ADLAR=1
        ADCSRA=0xE4;  //ADEN ADSC ADFR (ADIF)   (ADIE) ADPS2..0=1.0.0  (Division Factor=16)
   }



    g.tic++;
    if (40==g.tic){ //Anfang nchster Halbwelle 
      g.tic=0;
      g.hz10++;
      if (10==g.hz10){
        //Tachoauswertung - muss exakt alle 100ms aufgerufen werden.
        #define STEIGUNG 1   
        g.cMotorIst = TCNT0;
        TCNT0=0;
        
        if (g.cRegelung){
            //I-Regler fr Motor: neuen Stellwert fr Triac berechnen
            if ((g.cMotorIst < g.cMotorSoll) && (g.cZuendzeit >= STEIGUNG) ){
               g.cZuendzeit -=STEIGUNG;
            }else
        
            if ((g.cMotorIst > g.cMotorSoll) && (g.cZuendzeit<40-STEIGUNG) ){
               g.cZuendzeit +=STEIGUNG;
            }
        }
          
        g.hz10=0;  
        g.next100ms=1;
      }
        
      g.ms10++;
      if (100==g.ms10){
        g.ms10=0;  
        g.sec++;  
      }
    }          

    //Motortriac ggf. schalten
   if (g.tic > 38){ //Ende der Halbwelle. 
     PORTB &= (255-8); //Triac AUS.  clear PB3  
   }else{
     if (g.tic > g.cZuendzeit){
       PORTB |= 8;   //Triac EIN    set PB3
     }
   }    





}
//--------------------------------------------------------
ISR(TIMER1_CAPT_vect){ //Druckdosen Interrupt
    TCNT1 = 0;
    g.iDruckdose =ICR1;
    /*
  int16_t h,l;    
    l = ICR1L;
    h = ICR1H;
    g.iDruckdose = (h * 256 + l);
    */
}

//--------------------------------------------------------
void USART_Init(){
 
  g.rxIndexRead=0;
  g.rxIndexStore=0;
	/* Set baud rate: round(f_osc / (8*baudrate)) - 1  wenn U2X=1*/
	/* Set baud rate: round(f_osc / (16*baudrate)) - 1  wenn U2X=0*/
	UBRRH = 0;
	UBRRL = 25; //9600 Baud bei 4MHz

    UCSRA=0; //Clear U2X
    //UCSRA=2; //Set U2X

	/* Set frame format:          2stop bit   8data,*/
	UCSRC =           (1<<URSEL)|(1<<USBS)|(3<<UCSZ0);

	/* Enable Transmitter and  Receiver and RX-Interrupts (no TX-Interrupts)*/
	UCSRB = (1<<TXEN) | (1<< RXEN)  | (1<<RXCIE);
}


//--------------------------------------------------------
//--------------------------------------------------------
ISR(USART_RXC_vect){
  g.empfangszeichen = UDR;
  g.rxbuffer[g.rxIndexStore++] = g.empfangszeichen;
  if (g.rxIndexStore > 9) g.rxIndexStore=0;
}

//--------------------------------------------------------
char getRxChar(void){
  char c;
  c = g.rxbuffer[g.rxIndexRead++];
  if (g.rxIndexRead>9) g.rxIndexRead=0;
  return c;  

}

//--------------------------------------------------------
char receiveLine(void){
  static unsigned char n=9; //Am Ende starten. Startzeichen setzt auf Anfang.
  char zeichen;

  while (g.rxIndexStore != g.rxIndexRead) {
    zeichen = getRxChar();
    if ('*'==zeichen){
      n=0; //Zeilenanfang (nicht mit abspeichern)
      continue;
    }  
    if (n>9) n=9;  
    g.rxZeile[n] = zeichen;
    if (13==zeichen){
      g.rxZeile[n]=0; //String abschneiden  
      return 1; //Komplette Zeile
    }
    n++;
  }
  return 0; //Keine komplette Zeile
}
//--------------------------------------------------------
void USART_Transmit( unsigned char data ){
	/* Wait for empty transmit buffer */
	while ( !( UCSRA & (1<<UDRE)) )
	;
	/* Put data into buffer, sends the data */
	UDR = data;
}

//--------------------------------------------------------
void sendstring(unsigned char *text){
  while(*text){
    USART_Transmit(*text++);
  }
}


//--------------------------------------------------------
void Telegrammauswertung(void){
unsigned char *p;
   if (receiveLine()){
     switch (g.rxZeile[0]){
       case 'S': //Motorsolldrehzahl (Reglervorgabe)
         g.cMotorSoll = atoi(g.rxZeile+1);
/*
         p=stoa(g.txMsg, "S=");
         p=isTOa0(p, g.cMotorSoll,3,"\n\r");
         sendstring(g.txMsg);
*/
         break;

       case 's': //Motorstrom (direkte TRIAC-EIN-Zeit in 1/40 Halbwellenzeit)
         g.cZuendzeit = 40 - atoi(g.rxZeile+1);
/*
         p=stoa(g.txMsg, "S=");
         p=isTOa0(p, g.cMotorSoll,3,"\n\r");
         sendstring(g.txMsg);
*/
         break;

       case 'z': //Integrator istwert (cZuendzeit)
         p=stoa(g.txMsg, "z=");
         p=isTOa0(p, g.cZuendzeit,3,"\n\r");
         sendstring(g.txMsg);
         break;

       case 'R':  //Relais   ein     
         switch (g.rxZeile[1]){
           case 'L':PORTB |= 2; break;  //Links
           case 'R':PORTB |= 4; break;  //Rechts
           case 'S':PORTB |= 16; break; //Schnell
           case 'H':PORTB |= 32; break; //Heizung
         }             
         break;
         
       case 'r':  //Relais   aus        
         switch (g.rxZeile[1]){
           case 'L':PORTB &= (255-2); break;  //Links
           case 'R':PORTB &= (255-4); break;  //Rechts
           case 'S':PORTB &= (255-16); break; //Schnell
           case 'H':PORTB &= (255-32); break; //Heizung
         }             
         break;
         
       case 'T':   //Triac ein
         switch (g.rxZeile[1]){
           case 'P':PORTC |= 32; break;  //Pumpe
           case 'T':PORTD |= 8; break;   //Tuer
           case '1':PORTD |= 32; break;  //Ventil1
           case '2':PORTD |= 64; break;  //Ventil2
           case '3':PORTD |= 128; break; //Ventil3
         }             
         break;
         
       case 't':   //Triac aus
         switch (g.rxZeile[1]){
           case 'P':PORTC &= (255-32); break;  //Pumpe
           case 'T':PORTD &= (255-8); break;   //Tuer
           case '1':PORTD &= (255-32); break;  //Ventil1
           case '2':PORTD &= (255-64); break;  //Ventil2
           case '3':PORTD &= (255-128); break; //Ventil3
         }             
         break;
         
       case 'Q':   //Quiet
         g.talk = 0;
         break;

       case 'q':   //not Quiet
         g.talk = 1;
         break;

       case 'M':   //Motorregelung AUS
         g.cRegelung = 0;
         break;

       case 'm':   //Motorregelung EIN
         g.cRegelung = 1;
         break;

         
         
/*         
PB1     Links
PB2     Rechts
PB3     Triac Motor
PB4     Schnell
PB5     Heizung
PC5     Pumpe
PD1     Tx + Watchdog
PD3     Tuer
PD5     Ventil1
PD6     Ventil2
PD7     Ventil3
*/

     }
   }
}

//--------------------------------------------------------

int main (void)
{
int i;
int n;    
unsigned char *ram;


    cli();


    //AD-Wandler 
    ADMUX=0x60; //RefS1=0 RefS0=1 (Uref=AVcc) ADLAR=1  **  Bit 0..2=Channel
    ADCSRA=0x85;  //ADEN (ADSC) (ADFR) (ADIF) (ADIE) ADPS2..0=1.0.1  (Division Factor=32 => 125KHz ADC-Takt)

    //Set up the 8-bit timer 0 (Motordrehzahl)
    //-----------------------------------------
    TCCR0 =  7; //Drehgeber an T0 pin, Trigger auf rising Edge

    //Set up the 16-bit timer 1 (Wasserstand)
    //---------------------------------------
    //Dosen-Signal an ICP1 pin (Using the Input Capture Unit)
    
    TCCR1A =  0; //Normaler Timer/Counter
    TCCR1B =  _BV(ICNC1)  //Input Capture Noise Canceler
            | _BV(ICES1)  //Trigger on rising edge
            //_BV(CS12)
            | _BV(CS11) ; // Clock Select 010 => Teiler=8 (500KHz)
            //_BV(CS10)
            
            

    //Set up the 8-bit timer 2  (4KHz Generator)
    //------------------------------------------
    TCCR2 =  //_BV(FOC2)
            // _BV(WGM20)  
            //_BV(COM21)
            //_BV(COM20)
            _BV(WGM21)  //  CTC Modus
            
            //_BV(CS22)  
            | _BV(CS21)  // Clock Select 010 => Teiler=8 (500KHz)
            //| _BV(CS20) 
            //_BV(CS10)
            ;
            
    OCR2 = 124; //Zhlt 0..124, sollte genau 4KHz Interrupt-Frequenz ergeben

    TIMSK = _BV(TICIE1) //Timer 1 Input Capture
          | _BV(TOIE0)  //Timer 0 Overflow
          | _BV(OCIE2)  //Timer 2 Output Compare Match
          ;
    
    
    /*
    PB1     Links
    PB2     Rechts
    PB3     Triac Motor
    PB4     Schnell
    PB5     Heizung
    */
    DDRB = 2+4+8+16+32;
    PORTB = 0; //alles AUS

    /*
    PC5     Pumpe
    */
    DDRC = 32;
    PORTC = 0; //alles AUS
    
    /*
    (PD1     Tx + Watchdog) <-USART setzt DataDirection auf OUT
    PD3     Tuer
    PD5     Ventil1
    PD6     Ventil2
    PD7     Ventil3
    */ 
    DDRD = 8+32+64+128;
    PORTD = 0; //alles AUS

    USART_Init();

    //Globale Variablen initialisieren
//    memset(&g, 0, sizeof(Tglob));               Da rastet der Chip aus
    ram=(unsigned char *)&g;
    for (i=0; i< sizeof(Tglob); i++) ram[i]=0;  //Mach ich's halt von Hand
    g.cZuendzeit=40;  //entspricht AUS

    sei(); //Interrupts EIN. Jetzt geht's los!


    //Startmeldung senden. Derweil fhrt der MAX232 hoch.
    for (i=0; i<30; i++){
      sendstring("-- Controller startet...\n\r");
    }
    PORTD = 0; //alles AUS
   //Startmeldung
    sendstring("\n\r-- Waschmaschine\n\r");
    p=stoa(g.txMsg,"-- Vers. 0.3 vom ");  //   <---------------- Versionsnummer ----------
    p=stoa(p,__DATE__);
    p=stoa(p,"\n\r");
    sendstring(g.txMsg);

    n=0;

    g.talk=1;
    while(1){
      if (0==g.talk){
        if (g.empfangszeichen!=0){
          USART_Transmit(g.empfangszeichen); 
          g.empfangszeichen=0;
        }
      }
      

      Telegrammauswertung();
      if (g.next100ms){
        g.next100ms=0;  
        p=stoa(g.txMsg,"M");  //   (M)otor
        p=isTOa0(p, g.cMotorIst,3,"\n\r");
        if (g.talk){
            sendstring(g.txMsg);

            n++;
            if (n>4){  //alle 500ms
                n=0;  
                p=stoa(g.txMsg,"A");  //   (A)nalogwert
                p=uisTOa0(p, g.acAdcVal[0],3," ");
                p=uisTOa0(p, g.acAdcVal2[0],3," ");
                p=uisTOa0(p, g.acAdcVal[1],3," ");
                p=uisTOa0(p, g.acAdcVal2[1],3," ");
                p=uisTOa0(p, g.acAdcVal[2],3," ");
                p=uisTOa0(p, g.acAdcVal2[2],3," ");
                p=uisTOa0(p, g.acAdcVal[3],3," ");
                p=uisTOa0(p, g.acAdcVal2[3],3," ");
                p=uisTOa0(p, g.acAdcVal[4],3," ");
                p=uisTOa0(p, g.acAdcVal2[4],3," ");
                p=uisTOa0(p, g.cZuendzeit,2," ");
                p= isTOa0(p, g.iDruckdose,5,"\n\r");
              
                sendstring(g.txMsg);
            
            }
        }
      }
    }	
    return (0);
}
